#include <stdio.h>
#include <windows.h>

// from thp_play by thakis
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned __int64 u64;

typedef signed char s8;
typedef signed short s16;
typedef signed int s32;
typedef signed __int64 s64;

typedef float f32;


inline void toWORD(u16& w)
{
  u8 w1 = w & 0xff;
  u8 w2 = w >> 8;
  w = (w1 << 8) | w2;
}

inline void toSHORT(s16& w)
{
  toWORD(*(u16*)&w);
}

inline void toDWORD(u32& d)
{
  u8 w1 = d & 0xff;
  u8 w2 = (d >> 8) & 0xff;
  u8 w3 = (d >> 16) & 0xff;
  u8 w4 = d >> 24;
  d = (w1 << 24) | (w2 << 16) | (w3 << 8) | w4;
}

inline u32 aDWORD(u32 d)
{
  toDWORD(d); return d;
}

inline void toFLOAT(f32& f)
{ toDWORD(*(u32*)&f); }

struct ThpHeader
{
  char tag[4]; //'THP\0'

  //from monk's thp player:
  u32 version; //0x00011000 = 1.1, 0x00010000 = 1.0
  u32 maxBufferSize;
  u32 maxAudioSamples; //!= 0 if sound is stored in file


  float fps; //usually 29.something (=0x41efc28f) for ntsc
  u32 numFrames;
  u32 firstFrameSize; //size of first frame

  u32 dataSize; //size of file - ThpHeader.offset

  //from monk's thp player:
  u32 componentDataOffset; //ThpComponents stored here
  u32 offsetsDataOffset; //?? if != 0, offset to table with offsets of all frames?

  u32 firstFrameOffset;
  u32 lastFrameOffset;
};

//monk:

struct ThpComponents
{
  u32 numComponents; //usually 1 or 2 (video or video + audio)

  //component type 0 is video, type 1 is audio,
  //type 0xff is "no component" (numComponent many entries
  //are != 0xff)
  u8 componentTypes[16];
};

struct ThpVideoInfo
{
  u32 width;
  u32 height;
  u32 unknown; //only for version 1.1 thp files
};

struct ThpAudioInfo
{
  u32 numChannels;
  u32 frequency;
  u32 numSamples;
  u32 numData; //only for version 1.1 - that many
               //audio blocks are after each video block
               //(for surround sound?)
};

struct ThpAudioBlockHeader
{
  //size 80 byte
  u32 channelSize; //size of one channel in bytes
  u32 numSamples; //number of samples/channel
  s16 table1[16]; //table for first channel
  s16 table2[16]; //table for second channel
  s16 channel1Prev1;
  s16 channel1Prev2;
  s16 channel2Prev1;
  s16 channel2Prev2;
};

ThpHeader			_Head;
ThpComponents		_Components;
ThpVideoInfo		_VideoInfo;
ThpAudioInfo		_AudioInfo;
ThpAudioBlockHeader	_AudioBlockHeader;

int					_numInts;

u32				_currFrameNr;


void readThpHeader(HANDLE hFile, ThpHeader &h)
{
	DWORD	readBytes=0;

	ReadFile(hFile, &h, sizeof(ThpHeader) , &readBytes, NULL);

	if(readBytes==sizeof(ThpHeader))
	{
		toDWORD(h.version);  
		toDWORD(h.maxBufferSize);
		toDWORD(h.maxAudioSamples);
		toFLOAT(h.fps);
		toDWORD(h.numFrames);
		toDWORD(h.firstFrameSize);
		toDWORD(h.dataSize);
		toDWORD(h.componentDataOffset);
		toDWORD(h.offsetsDataOffset);
		toDWORD(h.firstFrameOffset);
		toDWORD(h.lastFrameOffset);
	}
}

void readThpComponents(HANDLE hFile, ThpComponents &c)
{
	DWORD	readBytes=0;

	ReadFile(hFile, &c, sizeof(ThpComponents) , &readBytes, NULL);

	toDWORD(c.numComponents);
}

void readThpVideoInfo(HANDLE hFile, ThpVideoInfo& i, bool isVersion11)
{
	DWORD	readBytes=0;
	
	ReadFile(hFile, &i, sizeof(ThpVideoInfo) , &readBytes, NULL);

	toDWORD(i.width);
	toDWORD(i.height);
	if(isVersion11)
		toDWORD(i.unknown);
	else
	{
		i.unknown = 0;
		SetFilePointer(hFile, -4, 0, FILE_CURRENT);
	}
}

void readThpAudioInfo(HANDLE hFile, ThpAudioInfo& i, bool isVersion11)
{
	DWORD	readBytes=0;

	ReadFile(hFile, &i, sizeof(ThpAudioInfo) , &readBytes, NULL);

	toDWORD(i.numChannels);
	toDWORD(i.frequency);
	toDWORD(i.numSamples);
	
	if(isVersion11)
		toDWORD(i.numData);
	else
	{
	    i.numData = 1;
		SetFilePointer(hFile, -4, 0, FILE_CURRENT);
	}	
}

int main(int argc, char* argv[])
{

	HANDLE	hFile;
	HANDLE	hFileSave;

	DWORD	readBytes=0;
	DWORD	writeBytes=0;

	u32		currAudioOffset;
	u32		nextAudioOffset;
	u32		offsetValue;

	u32		currAudioLength;
	u32		nextAudioLength;

	u32		currFrameOffset=0;
	u32		currFrameSize=0;
	u32		nextFrameOffset=0;
	u32		nextFrameSize=0;

	char	filename[MAX_PATH];
	char	fileout[MAX_PATH];

	bool	bSaveHeader=0;
	u8		Header[0x60];

	unsigned char *	_currFrameData;
	unsigned char *	_nextFrameData;
	
	if (argc!=2)
	{
		fprintf(stderr,"Usage : %s <filename>\n",argv[0]);
		ExitProcess(1);
	}

	strcpy_s(filename,MAX_PATH,argv[1]);
	strcpy_s(fileout,MAX_PATH,argv[1]);

	hFile=CreateFileA(filename,GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
	if(hFile==INVALID_HANDLE_VALUE)
	{
		fprintf(stderr,"Unable to open %s !\n",argv[1]);
		ExitProcess(1);
	}
	
	// Stock the THP Header
	ReadFile(hFile,Header,0x60,&readBytes,NULL);
	SetFilePointer(hFile,0,0,FILE_BEGIN);

	// Check if THP :P
	if(memcmp(Header,"THP\0x00",4))
	{
		fprintf(stdout,"%s is not a THP file ! \n",argv[1]);
		ExitProcess(1);
	}

	// Read Stuff ..
	readThpHeader(hFile,_Head);
	readThpComponents(hFile, _Components);

	for(u32 i = 0; i < _Components.numComponents; ++i)
	{
		if(_Components.componentTypes[i] == 0) //video
			readThpVideoInfo(hFile, _VideoInfo, _Head.version == 0x00011000);
		else if(_Components.componentTypes[i] == 1) //audio
		{
			readThpAudioInfo(hFile, _AudioInfo, _Head.version == 0x00011000);
		}
	}

	// Check if file has audio ...
	if(_Head.maxAudioSamples==0)
	{
		fprintf(stdout,"%s has no audio ! \n",argv[1]);
		ExitProcess(1);
	}

	memcpy(&fileout[strlen(fileout)-3],"DSP",3);

	hFileSave=CreateFileA(fileout,GENERIC_WRITE, FILE_APPEND_DATA,NULL,OPEN_ALWAYS,0,NULL);
	if(hFileSave==INVALID_HANDLE_VALUE)
	{
		fprintf(stderr,"Unable to open save file !\n",argv[1]);
		ExitProcess(1);
	}

	_currFrameNr = -1;
	nextFrameOffset = _Head.firstFrameOffset;
	nextFrameSize = _Head.firstFrameSize;

	_currFrameData = (unsigned char*)malloc(_Head.maxBufferSize);
	_nextFrameData = (unsigned char*)malloc(_Head.maxBufferSize);

	if(!(_currFrameData || _nextFrameData))
	{
		fprintf(stderr,"Unable to allocate %d bytes ! \n",_Head.maxBufferSize);
		ExitProcess(1);
	}
	
	// Display file info ...
	fprintf(stdout,"File : %s \n",filename);
	fprintf(stdout,"Out  : %s \n",fileout);
	if(_AudioInfo.numChannels==1)
		fprintf(stdout,"%d Hz mono\n\n",_AudioInfo.frequency);
	else
		fprintf(stdout,"%d Hz stereo\n\n",_AudioInfo.frequency);

	currFrameOffset=nextFrameOffset;
	
	do
	{
		++_currFrameNr;

		// Get the size of the first chunk ...
		SetFilePointer(hFile,nextFrameOffset,0,FILE_BEGIN);
		ReadFile(hFile, _currFrameData, nextFrameSize, &readBytes,NULL);
		
		currAudioOffset = aDWORD(*(u32*)(_currFrameData+0x08))+0x10;
		currAudioLength=aDWORD(*(u32*)(_currFrameData+currAudioOffset));
		currAudioLength*=_AudioInfo.numChannels;
		currAudioLength+=sizeof(ThpAudioBlockHeader)+0x010;

		currFrameOffset += nextFrameSize;
		currFrameSize = aDWORD(*(u32*)_currFrameData);
	
		// Get the size of the second chunk ...
		SetFilePointer(hFile,currFrameOffset,0,FILE_BEGIN);
		ReadFile(hFile, _nextFrameData, currFrameSize, &readBytes,NULL);

		nextAudioOffset = aDWORD(*(u32*)(_nextFrameData+0x08))+0x10;
		nextAudioLength = aDWORD(*(u32*)(_nextFrameData+nextAudioOffset));
		nextAudioLength *= _AudioInfo.numChannels;
		nextAudioLength+=sizeof(ThpAudioBlockHeader)+0x10;

		if(!bSaveHeader)
		{
			// Save the THPHeader
			bSaveHeader=1;
			offsetValue=aDWORD(currAudioLength);
		    memcpy(Header+0x18,&offsetValue,4);

			WriteFile(hFileSave,Header,0x60,&writeBytes,NULL);
		}

		// Write Offsets ...
		offsetValue=aDWORD(nextAudioLength);
		WriteFile(hFileSave,&offsetValue,4,&writeBytes,NULL);
		WriteFile(hFileSave,&offsetValue,4,&writeBytes,NULL);
		offsetValue=0;
		WriteFile(hFileSave,&offsetValue,4,&writeBytes,NULL);
		WriteFile(hFileSave,&offsetValue,4,&writeBytes,NULL);

		// Write Data
		WriteFile(hFileSave,_currFrameData+currAudioOffset,currAudioLength-0x10,&writeBytes,NULL);
		
		nextFrameOffset=currFrameOffset;
		nextFrameSize=currFrameSize;

		printf("Processing Frame %d on %d ...\r",_currFrameNr, _Head.numFrames+1);
		
	} while(_currFrameNr <= _Head.numFrames);
	
	CloseHandle(hFileSave);
	CloseHandle(hFile);
	free(_currFrameData);
	free(_nextFrameData);

	return 0;
}
